from emis_funky_funktions import *

from enum import auto, IntEnum
from functools import cache, reduce
from operator import getitem
from typing import Any, cast, Collection, Mapping, Sequence, Set, Tuple, TypeGuard

from lex import Tok

"""
Implements a parser for the following grammar:

Start    := PredicateSection <Idents> Newline
            VariablesSection <Idents> Newline
            ConstantsSection <Idents> Newline
            FunctionsSection <Idents> Newline
            ClausesSection <Clauses> Eof

Idents   := Identifier <Idents>
         := ε

Clauses  := Newline <Clauses'>
         := ε

Clauses' := <Clause> <Clauses>
         := ε

Clause   := <Term> <Clause'>

Clause'  := <Clause>
         := ε

Term     := Negate <Term>
         := Identifier <Func?>

Func?    := OpenP <Term> <CSTerms> CloseP
         := ε

CSTerms  := Comma <Term> <CSTerms>
         := ε
"""

class Variable(IntEnum):
    Start = auto()
    Idents = auto()
    Clauses = auto()
    Clauses_ = auto()
    Clause = auto()
    Clause_ = auto()
    Term = auto()
    Func = auto()
    CSTerms = auto()

    def __repr__(self) -> str:
        return f'<{self._name_}>'

GRAMMAR: Sequence[Tuple[Variable, Sequence[Variable | Tok]]] = [
    (Variable.Start,
         [ Tok.PredicateSection, Variable.Idents, Tok.Newline
         , Tok.VariablesSection, Variable.Idents, Tok.Newline
         , Tok.ConstantsSection, Variable.Idents, Tok.Newline
         , Tok.FunctionsSection, Variable.Idents, Tok.Newline
         , Tok.ClausesSection, Variable.Clauses, Tok.Eof ] ),

    (Variable.Idents,
        [ Tok.Identifier, Variable.Idents ]),
    (Variable.Idents,
        [ ]),

    (Variable.Clauses,
        [ Tok.Newline, Variable.Clauses_ ]),
    (Variable.Clauses,
        [ ]),

    (Variable.Clauses_,
        [ Variable.Clause, Variable.Clauses ]),
    (Variable.Clauses_,
        [ ]),

    (Variable.Clause,
        [ Variable.Term, Variable.Clause_ ]),

    (Variable.Clause_,
        [ Variable.Clause ]),
    (Variable.Clause_,
        [ ]),

    (Variable.Term,
        [ Tok.Negate, Variable.Term ]),
    (Variable.Term,
        [ Tok.Identifier, Variable.Func ]),

    (Variable.Func,
        [ Tok.OpenP, Variable.CSTerms, Tok.CloseP ]),
    (Variable.Func,
        [ ]),

    (Variable.CSTerms,
        [ Tok.Comma, Variable.Term, Variable.CSTerms ]),
    (Variable.CSTerms,
        [ ]),
]


# ### FIRST Table ###
#
# Start      : PredicateSection
# Idents     : Identifier, ε
# Clauses    : Newline, ε
# Clauses'   : Negate, Identifier, ε
# Clause     : Negate, Identifier
# Clause'    : Negate, Identifier, ε
# Term       : Negate, Identifier
# Func?      : OpenP
# CSTerms    : Comma, ε
#
#
#
# ### FOLLOW Table ###
#
# Idents     : Newline
# Clauses    : Eof
# Clauses'   : Eof
# Clause     : Newline, Eof
# Clause'    : Newline, Eof
# Term       : Negate, Identifier, Newline, Eof, Comma
# Func?      : Negate, Identifier, Newline, Eof, Comma
# CSTerms    : CloseP
#
#
#
# ### PREDICT Table ###
#
# Idents     : Identifier
#            : Newline
# Clauses    : Newline
#            : Eof
# Clauses'   : Negate, Identifier
#            : Eof
# Clause     : Newline, Eof
# Clause'    : Newline, Eof
# Term       : Negate, Identifier, Newline, Eof, Comma
# Func?      : Negate, Identifier, Newline, Eof, Comma
# CSTerms    : CloseP